# (C) Copyright 2018 by Rocky Bernstein
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
CPython 1.4 bytecode opcodes

This is used in bytecode disassembly. This is similar to the
opcodes in Python's dis.py library.
"""

# This is used from outside this module
from xdis.bytecode import findlabels

import xdis.opcodes.opcode_15 as opcode_15
from xdis.opcodes.base import (
    init_opdata,
    def_op, name_op, rm_op, varargs_op,
    finalize_opcodes, format_extended_arg,
    # Although these aren't used here, they are exported
    update_pj2
    )

version = 1.4

l = locals()
init_opdata(l, opcode_15, version)

# 1.4 Bytecodes not in 1.5
def_op(l, 'UNARY_CALL',         14)
def_op(l, 'BINARY_CALL',        26)
def_op(l, 'RAISE_EXCEPTION',    81)
def_op(l, 'BUILD_FUNCTION',     86)
varargs_op(l, 'UNPACK_ARG',     94)   # Number of arguments expected
varargs_op(l, 'UNPACK_VARARG',  99)  # Minimal number of arguments
name_op(l, 'LOAD_LOCAL',       115)
varargs_op(l, 'SET_FUNC_ARGS', 117)  # Argcount
varargs_op(l, 'RESERVE_FAST',  123)  # Number of local variables

update_pj2(globals(), l)

opcode_arg_fmt = {
    'EXTENDED_ARG': format_extended_arg,
}

finalize_opcodes(l)

def findlinestarts(co, dup_lines=False):
    code = co.co_code
    n = len(code)
    offset = 0
    while offset < n:
        op = code[offset]
        offset += 1
        if op == l['opmap']['SET_LINENO'] and offset > 0:
            lineno = code[offset] + code[offset+1]*256
            yield (offset-1, lineno)
            pass
        if op >= l['HAVE_ARGUMENT']:
            offset += 2
            pass
        pass



import os
import pytest
import re

from xdis.main import disassemble_file
from xdis import PYTHON3

if PYTHON3:
    from io import StringIO
else:
    from StringIO import StringIO

def get_srcdir():
    filename = os.path.normcase(os.path.dirname(__file__))
    return os.path.realpath(filename)


@pytest.mark.parametrize(("test_tuple", "function_to_test"), [
    (
        ('../test/bytecode_3.6/01_fstring.pyc', 'testdata/fstring-3.6.right',),
        disassemble_file,
    ),
    (
        ('../test/bytecode_3.0/04_raise.pyc', 'testdata/raise-3.0.right',),
        disassemble_file,
    ),
    (
        ('../test/bytecode_2.7pypy/04_pypy_lambda.pyc', 'testdata/pypy_lambda.right',),
        disassemble_file,
    ),
    (
        ('../test/bytecode_3.6/03_big_dict.pyc', 'testdata/big_dict-3.6.right',),
        disassemble_file,
    ),
])
def test_funcoutput(capfd, test_tuple, function_to_test):
    in_file, filename_expected = [
        os.path.join(get_srcdir(), p) for p in test_tuple
    ]
    resout = StringIO()
    function_to_test(in_file, resout)
    expected = "".join(open(filename_expected, "r").readlines())
    got_lines = resout.getvalue().split("\n")
    got_lines = [re.sub(' at 0x[0-9a-f]+', ' at 0xdeadbeef0000', line)
                 for line in got_lines]
    got_lines = [re.sub('<code object .*>|<xdis.code.Code[23] (object|instance) .*>',
                        '<xdis.code.thingy instance at 0xdeadbeef0000>',
                        line)
                 for line in got_lines]
    got = "\n".join(got_lines[5:])

    if 'XDIS_DONT_WRITE_DOT_GOT_FILES' not in os.environ:
      if got != expected:
          with open(filename_expected + ".got", "w") as out:
              out.write(got)
    assert got == expected



# std
import sys
from contextlib import closing
# compat
import six
# 3rd party
import pytest
# local
import xdis.std as dis
from xdis import PYTHON3, IS_PYPY, PYTHON_VERSION


# just a simple bit of code that should be the same across python versions,
# we are testing the api here really, leave it to the other tests to perform
# more complicated implementation verification.


TEST_SOURCE_CODE = 'a = 10'
TEST_CODE = compile(TEST_SOURCE_CODE, '<disassembly>', 'single')


TEST_BRANCH_SOURCE_CODE = 'a = 0 if 1 else 2'
TEST_BRANCH_CODE = compile(TEST_BRANCH_SOURCE_CODE, '<disassembly>', 'single')


EXPECTED_CODE_INFO = ("""# Method Name:       <module>
# Filename:          <disassembly>
# Argument count:    0
"""
+ ("# Kw-only arguments: 0\n" if PYTHON3 else "") +
"""# Number of locals:  0
# Stack size:        1
# Flags:             {flags}
# First Line:        1
# Constants:
#    0: 10
#    1: None
# Names:
#    0: a""").format(flags='0x00000000 (0x0)' if (
    IS_PYPY and PYTHON_VERSION < 3.5) else '0x00000040 (NOFREE)')

EXPECTED_DIS = """\
  1:           0 LOAD_CONST                0 (10)
               3 STORE_NAME                0 (a)
               6 LOAD_CONST                1 (None)
               9 RETURN_VALUE
"""

EXPECTED_DIS_36 = """\
  1:           0 LOAD_CONST                0 (10)
               2 STORE_NAME                0 (a)
               4 LOAD_CONST                1 (None)
               6 RETURN_VALUE
"""

if PYTHON_VERSION < 3.6:
    expected_dis = EXPECTED_DIS
else:
    expected_dis = EXPECTED_DIS_36


@pytest.fixture
def bytecode_fixture():
    return dis.Bytecode(TEST_SOURCE_CODE)


@pytest.fixture
def traceback_fixture():
    try:
        raise Exception
    except:
        _, _, tb = sys.exc_info()
        return tb


@pytest.yield_fixture
def stream_fixture():
    with closing(six.StringIO()) as s:
        yield s


def test_bytecode_from_traceback(traceback_fixture):
    assert len(list(dis.Bytecode.from_traceback(traceback_fixture))) > 0


def test_bytecode_codeobj(bytecode_fixture):
    assert bytecode_fixture.codeobj is not None


def test_bytecode_first_line(bytecode_fixture):
    assert bytecode_fixture.first_line is not None


def test_bytecode_dis(bytecode_fixture):
    assert bytecode_fixture.dis() == expected_dis


def test_bytecode_info(bytecode_fixture):
    actual = bytecode_fixture.info()
    assert actual == EXPECTED_CODE_INFO


def test_bytecode__iter__(bytecode_fixture):
    assert len(list(bytecode_fixture)) > 0


def test_code_info():
    assert dis.code_info(TEST_SOURCE_CODE) == EXPECTED_CODE_INFO


def test_show_code(stream_fixture):
    dis.show_code(TEST_SOURCE_CODE, file=stream_fixture)
    actual = stream_fixture.getvalue()
    assert actual == EXPECTED_CODE_INFO + '\n'


def test_pretty_flags():
    assert dis.pretty_flags(1) == '0x00000001 (OPTIMIZED)'


def test_dis(stream_fixture):
    dis.dis(TEST_SOURCE_CODE, file=stream_fixture)
    actual = stream_fixture.getvalue()
    assert actual == expected_dis + '\n'


def test_distb(traceback_fixture, stream_fixture):
    dis.distb(traceback_fixture, file=stream_fixture)
    actual = stream_fixture.getvalue()
    actual_len = len(actual)
    assert actual_len > 0


def test_disassemble(stream_fixture):
    dis.disassemble(TEST_CODE, file=stream_fixture)
    actual = stream_fixture.getvalue()
    expected = EXPECTED_CODE_INFO + '\n' + expected_dis + '\n'
    assert actual == expected


def test_get_instructions():
    actual = list(dis.get_instructions(TEST_SOURCE_CODE))
    actual_len = len(actual)
    assert actual_len > 0


def test_findlinestarts():
    actual = list(dis.findlinestarts(TEST_CODE))
    actual_len = len(actual)
    assert actual_len > 0


def test_findlabels():
    if PYTHON_VERSION < 3.6:
        test_code = TEST_BRANCH_CODE
    else:
        test_code = TEST_BRANCH_CODE.co_code
    actual = list(dis.findlabels(test_code))
    actual_len = len(actual)
    assert actual_len > 0
